@@ -26,6 +26,7 @@ from sqlalchemy import orm
 from sqlalchemy.orm import exc as orm_exc
 from sqlalchemy import schema
 
+from tacker._i18n import _
 from tacker.api.v1 import attributes
 from tacker.common import exceptions
 from tacker import context as t_context
@@ -40,10 +41,12 @@ from tacker import manager
 from tacker.plugins.common import constants
 
 LOG = logging.getLogger(__name__)
-_ACTIVE_UPDATE = (constants.ACTIVE, constants.PENDING_UPDATE)
+_ACTIVE_UPDATE = (constants.ACTIVE, constants.PENDING_UPDATE,
+                  constants.PENDING_HEAL)
 _ACTIVE_UPDATE_ERROR_DEAD = (
     constants.PENDING_CREATE, constants.ACTIVE, constants.PENDING_UPDATE,
-    constants.ERROR, constants.DEAD)
+    constants.PENDING_SCALE_IN, constants.PENDING_SCALE_OUT, constants.ERROR,
+    constants.PENDING_DELETE, constants.DEAD, constants.PENDING_HEAL)
 CREATE_STATES = (constants.PENDING_CREATE, constants.DEAD)
 
 
@@ -129,8 +132,8 @@ class VNF(model_base.BASE, models_v1.HasId, models_v1.HasTenant,
 
     # For a management tool to talk to manage this hosting vnf.
     # opaque string.
-    # e.g. (driver, mgmt_url) = (ssh, ip address), ...
-    mgmt_url = sa.Column(sa.String(255), nullable=True)
+    # e.g. (driver, mgmt_ip_address) = (ssh, ip address), ...
+    mgmt_ip_address = sa.Column(sa.String(255), nullable=True)
     attributes = orm.relationship("VNFAttribute", backref="vnf")
 
     status = sa.Column(sa.String(64), nullable=False)
@@ -225,7 +228,8 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
         }
         key_list = ('id', 'tenant_id', 'name', 'description', 'instance_id',
                     'vim_id', 'placement_attr', 'vnfd_id', 'status',
-                    'mgmt_url', 'error_reason', 'created_at', 'updated_at')
+                    'mgmt_ip_address', 'error_reason', 'created_at',
+                    'updated_at')
         res.update((key, vnf_db[key]) for key in key_list)
         return self._fields(res, fields)
 
@@ -339,6 +343,8 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
 
     def get_vnfd(self, context, vnfd_id, fields=None):
         vnfd_db = self._get_resource(context, VNFD, vnfd_id)
+        if not vnfd_db:
+            raise exceptions.NotFound(resource='VNFD', name=vnfd_id)
         return self._make_vnfd_dict(vnfd_db)
 
     def get_vnfds(self, context, filters, fields=None):
@@ -434,14 +440,15 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
     # called internally, not by REST API
     # intsance_id = None means error on creation
     def _create_vnf_post(self, context, vnf_id, instance_id,
-                         mgmt_url, vnf_dict):
+                         mgmt_ip_address, vnf_dict):
         LOG.debug('vnf_dict %s', vnf_dict)
         with context.session.begin(subtransactions=True):
             query = (self._model_query(context, VNF).
                      filter(VNF.id == vnf_id).
                      filter(VNF.status.in_(CREATE_STATES)).
                      one())
-            query.update({'instance_id': instance_id, 'mgmt_url': mgmt_url})
+            query.update({'instance_id': instance_id,
+                          'mgmt_ip_address': mgmt_ip_address})
             if instance_id is None or vnf_dict['status'] == constants.ERROR:
                 query.update({'status': constants.ERROR})
 
@@ -451,7 +458,8 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
                     self._vnf_attribute_update_or_create(context, vnf_id,
                                                          key, value)
         evt_details = ("Infra Instance ID created: %s and "
-                       "Mgmt URL set: %s") % (instance_id, mgmt_url)
+                       "Mgmt IP address set: %s") % (instance_id,
+                                                     mgmt_ip_address)
         self._cos_db_plg.create_event(
             context, res_id=vnf_dict['id'],
             res_type=constants.RES_TYPE_VNF,
@@ -472,7 +480,7 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
                 evt_type=constants.RES_EVT_CREATE,
                 tstamp=timeutils.utcnow(), details="VNF creation completed")
 
-    def _get_vnf_db(self, context, vnf_id, current_statuses, new_status):
+    def _get_vnf_db(self, context, vnf_id, current_statuses):
         try:
             vnf_db = (
                 self._model_query(context, VNF).
@@ -481,22 +489,44 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
                 with_lockmode('update').one())
         except orm_exc.NoResultFound:
             raise vnfm.VNFNotFound(vnf_id=vnf_id)
-        if vnf_db.status == constants.PENDING_UPDATE:
-            raise vnfm.VNFInUse(vnf_id=vnf_id)
+        return vnf_db
+
+    def _update_vnf_status_db(self, context, vnf_id, current_statuses,
+                              new_status):
+        vnf_db = self._get_vnf_db(context, vnf_id, current_statuses)
+        if self.check_vnf_status_legality(vnf_db, vnf_id):
+            vnf_db.update({'status': new_status})
+        return vnf_db
+
+    def _update_vnf_status_db_no_check(self, context, vnf_id, current_statuses,
+                              new_status):
+        vnf_db = self._get_vnf_db(context, vnf_id, current_statuses)
         vnf_db.update({'status': new_status})
         return vnf_db
 
+    @staticmethod
+    def check_vnf_status_legality(vnf_db, vnf_id):
+        if vnf_db.status == constants.PENDING_DELETE:
+            error_reason = _("Operation on PENDING_DELETE VNF "
+                             "is not permited. Please contact your "
+                             "Administrator.")
+            raise vnfm.VNFDeleteFailed(reason=error_reason)
+        if(vnf_db.status in [constants.PENDING_UPDATE,
+                             constants.PENDING_HEAL]):
+            raise vnfm.VNFInUse(vnf_id=vnf_id)
+        return True
+
     def _update_vnf_scaling_status(self,
                                    context,
                                    policy,
                                    previous_statuses,
                                    status,
-                                   mgmt_url=None):
+                                   mgmt_ip_address=None):
         with context.session.begin(subtransactions=True):
-            vnf_db = self._get_vnf_db(
+            vnf_db = self._update_vnf_status_db(
                 context, policy['vnf']['id'], previous_statuses, status)
-            if mgmt_url:
-                vnf_db.update({'mgmt_url': mgmt_url})
+            if mgmt_ip_address:
+                vnf_db.update({'mgmt_ip_address': mgmt_ip_address})
         updated_vnf_dict = self._make_vnf_dict(vnf_db)
         self._cos_db_plg.create_event(
             context, res_id=updated_vnf_dict['id'],
@@ -506,28 +536,30 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
             tstamp=timeutils.utcnow())
         return updated_vnf_dict
 
-    def _update_vnf_pre(self, context, vnf_id):
+    def _update_vnf_pre(self, context, vnf_id, new_status):
         with context.session.begin(subtransactions=True):
-            vnf_db = self._get_vnf_db(
-                context, vnf_id, _ACTIVE_UPDATE, constants.PENDING_UPDATE)
+            vnf_db = self._update_vnf_status_db(
+                context, vnf_id, _ACTIVE_UPDATE, new_status)
         updated_vnf_dict = self._make_vnf_dict(vnf_db)
-        self._cos_db_plg.create_event(
-            context, res_id=vnf_id,
-            res_type=constants.RES_TYPE_VNF,
-            res_state=updated_vnf_dict['status'],
-            evt_type=constants.RES_EVT_UPDATE,
-            tstamp=timeutils.utcnow())
+        if new_status in constants.VNF_STATUS_TO_EVT_TYPES:
+            self._cos_db_plg.create_event(
+                context, res_id=vnf_id,
+                res_type=constants.RES_TYPE_VNF,
+                res_state=updated_vnf_dict['status'],
+                evt_type=constants.VNF_STATUS_TO_EVT_TYPES[new_status],
+                tstamp=timeutils.utcnow())
         return updated_vnf_dict
 
     def _update_vnf_post(self, context, vnf_id, new_status,
-                         new_vnf_dict):
+                         new_vnf_dict, vnf_status, evt_type):
         updated_time_stamp = timeutils.utcnow()
         with context.session.begin(subtransactions=True):
             (self._model_query(context, VNF).
              filter(VNF.id == vnf_id).
-             filter(VNF.status == constants.PENDING_UPDATE).
+             filter(VNF.status == vnf_status).
              update({'status': new_status,
-                     'updated_at': updated_time_stamp}))
+                     'updated_at': updated_time_stamp,
+                     'mgmt_ip_address': new_vnf_dict['mgmt_ip_address']}))
 
             dev_attrs = new_vnf_dict.get('attributes', {})
             (context.session.query(VNFAttribute).
@@ -543,37 +575,59 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
             context, res_id=vnf_id,
             res_type=constants.RES_TYPE_VNF,
             res_state=new_status,
-            evt_type=constants.RES_EVT_UPDATE,
+            evt_type=evt_type,
             tstamp=updated_time_stamp)
 
-    def _delete_vnf_pre(self, context, vnf_id):
+    def _delete_vnf_pre(self, context, vnf_id, force_delete=False):
         with context.session.begin(subtransactions=True):
 
             nss_db = context.session.query(ns_db.NS).filter(
                 ns_db.NS.vnf_ids.like("%" + vnf_id + "%")).first()
-            if (nss_db is not None and nss_db.status not in
-                    [constants.PENDING_DELETE, constants.ERROR]):
-                raise vnfm.VNFInUse(vnf_id=vnf_id)
 
-            vnf_db = self._get_vnf_db(
-                context, vnf_id, _ACTIVE_UPDATE_ERROR_DEAD,
+            if not force_delete:
+                if (nss_db is not None and nss_db.status not in
+                        [constants.PENDING_DELETE, constants.ERROR]):
+                    raise vnfm.VNFInUse(vnf_id=vnf_id)
+
+                vnf_db = self._update_vnf_status_db(
+                    context, vnf_id, _ACTIVE_UPDATE_ERROR_DEAD,
+                    constants.PENDING_DELETE)
+            else:
+                vnf_db = self._update_vnf_status_db_no_check(context,
+                    vnf_id, _ACTIVE_UPDATE_ERROR_DEAD,
                 constants.PENDING_DELETE)
         deleted_vnf_db = self._make_vnf_dict(vnf_db)
+        details = "VNF delete initiated" if not force_delete else \
+            "VNF force delete initiated"
         self._cos_db_plg.create_event(
             context, res_id=vnf_id,
             res_type=constants.RES_TYPE_VNF,
             res_state=deleted_vnf_db['status'],
             evt_type=constants.RES_EVT_DELETE,
-            tstamp=timeutils.utcnow(), details="VNF delete initiated")
+            tstamp=timeutils.utcnow(), details=details)
         return deleted_vnf_db
 
-    def _delete_vnf_post(self, context, vnf_dict, error, soft_delete=True):
+    def _delete_vnf_force(self, context, vnf_id):
+        # Check mapping vnf in vnffg_db
+        with context.session.begin(subtransactions=True):
+            nss_db = context.session.query(ns_db.NS).filter(
+                ns_db.NS.vnf_ids.like("%" + vnf_id + "%")).first()
+            if nss_db:
+                pass
+
+    def _delete_vnf_post(self, context, vnf_dict, error,
+                         soft_delete=True, force_delete=False):
         vnf_id = vnf_dict['id']
         with context.session.begin(subtransactions=True):
-            query = (
-                self._model_query(context, VNF).
-                filter(VNF.id == vnf_id).
-                filter(VNF.status == constants.PENDING_DELETE))
+            if force_delete:
+                query = (
+                    self._model_query(context, VNF).
+                    filter(VNF.id == vnf_id))
+            else:
+                query = (
+                    self._model_query(context, VNF).
+                    filter(VNF.id == vnf_id).
+                    filter(VNF.status == constants.PENDING_DELETE))
             if error:
                 query.update({'status': constants.ERROR})
                 self._cos_db_plg.create_event(
@@ -619,7 +673,8 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
 
     # reference implementation. needs to be overrided by subclass
     def update_vnf(self, context, vnf_id, vnf):
-        vnf_dict = self._update_vnf_pre(context, vnf_id)
+        new_status = constants.PENDING_UPDATE
+        vnf_dict = self._update_vnf_pre(context, vnf_id, new_status)
         # start actual update of hosting vnf
         # waiting for completion of update should be done backgroundly
         # by another thread if it takes a while
@@ -681,11 +736,9 @@ class VNFMPluginDb(vnfm.VNFMPluginBase, db_base.CommonDbMixin):
 
     def _mark_vnf_dead(self, vnf_id):
         exclude_status = [
-            constants.DOWN,
             constants.PENDING_CREATE,
             constants.PENDING_UPDATE,
             constants.PENDING_DELETE,
-            constants.INACTIVE,
             constants.ERROR]
         return self._mark_vnf_status(
             vnf_id, exclude_status, constants.DEAD)
