@@ -17,6 +17,7 @@
 Utility methods for working with WSGI servers
 """
 from __future__ import print_function
+import functools
 
 import errno
 import os
@@ -40,6 +41,7 @@ import six
 import webob.dec
 import webob.exc
 
+from tacker._i18n import _
 from tacker.common import exceptions as exception
 from tacker import context
 from tacker.db import api
@@ -92,6 +94,54 @@ def encode_body(body):
     return encodeutils.to_utf8(body)
 
 
+def expected_errors(errors):
+    """Decorator for Restful API methods which specifies expected exceptions.
+
+    Specify which exceptions may occur when an API method is called. If an
+    unexpected exception occurs then return a 500 instead and ask the user
+    of the API to file a bug report.
+    """
+    def decorator(f):
+        @functools.wraps(f)
+        def wrapped(*args, **kwargs):
+            try:
+                return f(*args, **kwargs)
+            except Exception as exc:
+                if isinstance(exc, webob.exc.WSGIHTTPException):
+                    if isinstance(errors, int):
+                        t_errors = (errors,)
+                    else:
+                        t_errors = errors
+                    if exc.code in t_errors:
+                        raise
+                elif isinstance(exc, exception.Forbidden):
+                    # Note(nirajsingh): Special case to handle
+                    # Forbidden exceptions so every
+                    # extension method does not need to wrap authorize
+                    # calls. ResourceExceptionHandler silently
+                    # converts NotAuthorized to HTTPForbidden
+                    raise
+                elif isinstance(exc, exception.ValidationError):
+                    # Note(nirajsingh): Handle a validation error, which
+                    # happens due to invalid API parameters, as an
+                    # expected error.
+                    raise
+                elif isinstance(exc, exception.NotAuthorized):
+                    # Handle an authorized exception, will be
+                    # automatically converted to a HTTP 401.
+                    raise
+
+                LOG.exception("Unexpected exception in API method")
+                msg = _('Unexpected API Error. Please report this at '
+                    'http://bugs.launchpad.net/tacker/ and attach the Tacker '
+                    'API log if possible.\n%s') % type(exc)
+                raise webob.exc.HTTPInternalServerError(explanation=msg)
+
+        return wrapped
+
+    return decorator
+
+
 class WorkerService(common_service.ServiceBase):
     """Wraps a worker to be handled by ProcessLauncher."""
 
@@ -353,7 +403,7 @@ class Request(webob.Request):
         return bm or 'application/json'
 
     def get_content_type(self):
-        allowed_types = ("application/json")
+        allowed_types = ("application/json", "application/zip")
         if "Content-Type" not in self.headers:
             LOG.debug("Missing Content-Type")
             return None
@@ -409,7 +459,7 @@ class JSONDictSerializer(DictSerializer):
     def default(self, data):
         def sanitizer(obj):
             return six.text_type(obj)
-        return encode_body(jsonutils.dumps(data, default=sanitizer))
+        return encode_body(jsonutils.dump_as_bytes(data, default=sanitizer))
 
 
 class ResponseHeaderSerializer(ActionDispatcher):
@@ -428,6 +478,7 @@ class ResponseSerializer(object):
     def __init__(self, body_serializers=None, headers_serializer=None):
         self.body_serializers = {
             'application/json': JSONDictSerializer(),
+            'application/zip': JSONDictSerializer()
         }
         self.body_serializers.update(body_serializers or {})
 
@@ -485,6 +536,15 @@ class JSONDeserializer(TextDeserializer):
         return {'body': self._from_json(datastring)}
 
 
+class ZipDeserializer(ActionDispatcher):
+
+    def deserialize(self, body_file, action='default'):
+        return self.dispatch(body_file, action=action)
+
+    def default(self, body_file):
+        return {'body': body_file}
+
+
 class RequestHeadersDeserializer(ActionDispatcher):
     """Default request headers deserializer."""
 
@@ -551,7 +611,12 @@ class RequestDeserializer(object):
                 LOG.debug("Unable to deserialize body as provided "
                           "Content-Type")
 
-        return deserializer.deserialize(request.body, action)
+        if isinstance(deserializer, ZipDeserializer):
+            body = request.body_file
+        else:
+            body = request.body
+
+        return deserializer.deserialize(body, action)
 
     def get_body_deserializer(self, content_type):
         try:
@@ -682,6 +747,33 @@ class Debug(Middleware):
         print()
 
 
+class DefaultMethodController(object):
+    """Controller that handles the OPTIONS request method.
+
+    This controller handles the OPTIONS request method and any of the HTTP
+    methods that are not explicitly implemented by the application.
+    """
+
+    def options(self, request, **kwargs):
+        """Return a response that includes the 'Allow' header.
+
+        Return a response that includes the 'Allow' header listing the methods
+        that are implemented. A 204 status code is used for this response.
+        """
+        headers = [('Allow', kwargs.get('allowed_methods'))]
+        raise webob.exc.HTTPNoContent(headers=headers)
+
+    def reject(self, request, **kwargs):
+        """Return a 405 method not allowed error.
+
+        As a convenience, the 'Allow' header with the list of implemented
+        methods is included in the response as well.
+        """
+        headers = [('Allow', kwargs.get('allowed_methods'))]
+        raise webob.exc.HTTPMethodNotAllowed(
+            headers=headers)
+
+
 class Router(object):
     """WSGI middleware that maps incoming requests to WSGI apps."""
 
@@ -714,6 +806,7 @@ class Router(object):
           mapper.connect(None, "/v1.0/{path_info:.*}", controller=BlogApp())
         """
         self.map = mapper
+        self._setup_routes(self.map)
         self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                           self.map)
 
@@ -743,6 +836,146 @@ class Router(object):
         app = match['controller']
         return app
 
+    def _setup_routes(self, mapper):
+        pass
+
+
+class ResourceExceptionHandler(object):
+    """Context manager to handle Resource exceptions.
+
+    Used when processing exceptions generated by API implementation
+    methods.  Converts most exceptions to Fault
+    exceptions, with the appropriate logging.
+    """
+
+    def __enter__(self):
+        return None
+
+    def __exit__(self, ex_type, ex_value, ex_traceback):
+        if not ex_value:
+            return True
+        if isinstance(ex_value, exception.Forbidden):
+            raise Fault(webob.exc.HTTPForbidden(
+                explanation=ex_value.format_message()))
+        elif isinstance(ex_value, exception.BadRequest):
+            raise Fault(exception.ConvertedException(
+                code=ex_value.code,
+                explanation=ex_value.format_message()))
+        elif isinstance(ex_value, TypeError):
+            exc_info = (ex_type, ex_value, ex_traceback)
+            LOG.error('Exception handling resource: %s', ex_value,
+                      exc_info=exc_info)
+            raise Fault(webob.exc.HTTPBadRequest())
+        elif isinstance(ex_value, Fault):
+            LOG.error("Fault thrown: %s", ex_value)
+            raise ex_value
+        elif isinstance(ex_value, webob.exc.HTTPException):
+            LOG.error("HTTP exception thrown: %s", ex_value)
+            raise Fault(ex_value)
+
+        # We didn't handle the exception
+        return False
+
+
+def response(code):
+    """Attaches response code to a method.
+
+    This decorator associates a response code with a method.  Note
+    that the function attributes are directly manipulated; the method
+    is not wrapped.
+    """
+
+    def decorator(func):
+        func.wsgi_code = code
+        return func
+
+    return decorator
+
+
+class ResponseObject(object):
+    """Bundles a response object
+
+    Object that app methods may return in order to allow its response
+    to be modified by extensions in the code. Its use is optional (and
+    should only be used if you really know what you are doing).
+    """
+
+    def __init__(self, obj, code=None, headers=None):
+        """Builds a response object."""
+
+        self.obj = obj
+        self._default_code = 200
+        self._code = code
+        self._headers = headers or {}
+        self.serializer = JSONDictSerializer()
+
+    def __getitem__(self, key):
+        """Retrieves a header with the given name."""
+
+        return self._headers[key.lower()]
+
+    def __setitem__(self, key, value):
+        """Sets a header with the given name to the given value."""
+
+        self._headers[key.lower()] = value
+
+    def __delitem__(self, key):
+        """Deletes the header with the given name."""
+
+        del self._headers[key.lower()]
+
+    def serialize(self, request, content_type):
+        """Serializes the wrapped object.
+
+        Utility method for serializing the wrapped object.  Returns a
+        webob.Response object.
+
+        Header values are set to the appropriate Python type and
+        encoding demanded by PEP 3333: whatever the native str type is.
+        """
+
+        serializer = self.serializer
+
+        body = None
+        if self.obj is not None:
+            body = serializer.serialize(self.obj)
+        response = webob.Response(body=body)
+        response.status_int = self.code
+        for hdr, val in self._headers.items():
+            if six.PY2:
+                # In Py2.X Headers must be a UTF-8 encode str.
+                response.headers[hdr] = encodeutils.safe_encode(val)
+            else:
+                # In Py3.X Headers must be a str that was first safely
+                # encoded to UTF-8 (to catch any bad encodings) and then
+                # decoded back to a native str.
+                response.headers[hdr] = encodeutils.safe_decode(
+                    encodeutils.safe_encode(val))
+        # Deal with content_type
+        if not isinstance(content_type, six.text_type):
+            content_type = six.text_type(content_type)
+        if six.PY2:
+            # In Py2.X Headers must be a UTF-8 encode str.
+            response.headers['Content-Type'] = encodeutils.safe_encode(
+                content_type)
+        else:
+            # In Py3.X Headers must be a str.
+            response.headers['Content-Type'] = encodeutils.safe_decode(
+                encodeutils.safe_encode(content_type))
+        return response
+
+    @property
+    def code(self):
+        """Retrieve the response status."""
+
+        return self._code or self._default_code
+
+    @property
+    def headers(self):
+        """Retrieve the headers."""
+
+        return self._headers.copy()
+
 
 class Resource(Application):
     """WSGI app that handles (de)serialization and controller dispatch.
@@ -757,8 +990,7 @@ class Resource(Application):
 
     """
 
-    def __init__(self, controller, fault_body_function,
-                 deserializer=None, serializer=None):
+    def __init__(self, controller, deserializer=None, serializer=None):
         """Object initialization.
 
         :param controller: object that implement methods created by routes lib
@@ -766,15 +998,10 @@ class Resource(Application):
                              controller into a webob response
         :param serializer: object that can deserialize a webob request
                            into necessary pieces
-        :param fault_body_function: a function that will build the response
-                                    body for HTTP errors raised by operations
-                                    on this resource object
-
         """
         self.controller = controller
         self.deserializer = deserializer or RequestDeserializer()
         self.serializer = serializer or ResponseSerializer()
-        self._fault_body_function = fault_body_function
 
     @webob.dec.wsgify(RequestClass=Request)
     def __call__(self, request):
@@ -787,31 +1014,40 @@ class Resource(Application):
             action, args, accept = self.deserializer.deserialize(request)
         except exception.InvalidContentType:
             LOG.exception("InvalidContentType: Unsupported Content-Type")
-            return Fault(webob.exc.HTTPBadRequest(
+            return Fault(webob.exc.HTTPUnsupportedMediaType(
                 explanation=_("Unsupported Content-Type")))
         except exception.MalformedRequestBody:
             LOG.exception("MalformedRequestBody: Malformed request body")
             return Fault(webob.exc.HTTPBadRequest(
                 explanation=_("Malformed request body")))
 
+        response = None
         try:
-            action_result = self.dispatch(request, action, args)
-        except webob.exc.HTTPException as ex:
-            LOG.info("HTTP exception thrown: %s", ex)
-            action_result = Fault(ex,
-                                  self._fault_body_function)
-        except Exception:
-            LOG.exception("Internal error")
-            # Do not include the traceback to avoid returning it to clients.
-            action_result = Fault(webob.exc.HTTPServerError(),
-                                  self._fault_body_function)
+            with ResourceExceptionHandler():
+                action_result = self.dispatch(request, action, args)
+        except Fault as ex:
+            response = ex
+        except Exception as ex:
+            raise Fault(webob.exc.HTTPInternalServerError())
 
-        if isinstance(action_result, dict) or action_result is None:
-            response = self.serializer.serialize(action_result,
-                                                 accept,
-                                                 action=action)
-        else:
-            response = action_result
+        if not response:
+            resp_obj = None
+            if type(action_result) is dict or action_result is None:
+                resp_obj = ResponseObject(action_result)
+            elif isinstance(action_result, ResponseObject):
+                resp_obj = action_result
+            else:
+                response = action_result
+
+            # Run post-processing extensions
+            if resp_obj:
+                method = getattr(self.controller, action)
+                # Do a preserialize to set up the response object
+                if hasattr(method, 'wsgi_code'):
+                    resp_obj._default_code = method.wsgi_code
+
+            if resp_obj and not response:
+                response = resp_obj.serialize(request, accept)
 
         try:
             msg_dict = dict(url=request.url, status=response.status_int)
@@ -828,13 +1064,7 @@ class Resource(Application):
         """Find action-spefic method on controller and call it."""
 
         controller_method = getattr(self.controller, action)
-        try:
-            # NOTE(salvatore-orlando): the controller method must have
-            # an argument whose name is 'request'
-            return controller_method(request=request, **action_args)
-        except TypeError as exc:
-            LOG.exception(exc)
-            return Fault(webob.exc.HTTPBadRequest())
+        return controller_method(request=request, **action_args)
 
 
 def _default_body_function(wrapped_exc):
@@ -849,28 +1079,62 @@ def _default_body_function(wrapped_exc):
 
 
 class Fault(webob.exc.HTTPException):
-    """Generates an HTTP response from a webob HTTP exception."""
+    """Wrap webob.exc.HTTPException to provide API friendly response."""
 
-    def __init__(self, exception, body_function=None):
-        """Creates a Fault for the given webob.exc.exception."""
+    _fault_names = {
+        400: "badRequest",
+        401: "unauthorized",
+        403: "forbidden",
+        404: "itemNotFound",
+        405: "badMethod",
+        409: "conflictingRequest",
+        413: "overLimit",
+        415: "badMediaType",
+        429: "overLimit",
+        501: "notImplemented",
+        503: "serviceUnavailable"}
+
+    def __init__(self, exception):
+        """Create a Fault for the given webob.exc.exception."""
         self.wrapped_exc = exception
-        self.status_int = self.wrapped_exc.status_int
-        self._body_function = body_function or _default_body_function
+        for key, value in list(self.wrapped_exc.headers.items()):
+            self.wrapped_exc.headers[key] = str(value)
+        self.status_int = exception.status_int
 
     @webob.dec.wsgify(RequestClass=Request)
     def __call__(self, req):
         """Generate a WSGI response based on the exception passed to ctor."""
+        user_locale = req.best_match_language()
         # Replace the body with fault details.
-        fault_data, metadata = self._body_function(self.wrapped_exc)
-        content_type = req.best_match_content_type()
-        serializer = {
-            'application/json': JSONDictSerializer(),
-        }[content_type]
+        code = self.wrapped_exc.status_int
+        fault_name = self._fault_names.get(code, "tackerFault")
+        explanation = self.wrapped_exc.explanation
+        LOG.debug("Returning %(code)s to user: %(explanation)s",
+                  {'code': code, 'explanation': explanation})
+
+        explanation = i18n.translate(explanation, user_locale)
+        fault_data = {
+            fault_name: {
+                'code': code,
+                'message': explanation}}
+        if code == 413 or code == 429:
+            retry = self.wrapped_exc.headers.get('Retry-After', None)
+            if retry:
+                fault_data[fault_name]['retryAfter'] = retry
+
+        self.wrapped_exc.content_type = 'application/json'
+        self.wrapped_exc.charset = 'UTF-8'
+
+        body = JSONDictSerializer().serialize(fault_data)
+        if isinstance(body, six.text_type):
+            body = body.encode('utf-8')
+        self.wrapped_exc.body = body
 
-        self.wrapped_exc.body = serializer.serialize(fault_data)
-        self.wrapped_exc.content_type = content_type
         return self.wrapped_exc
 
+    def __str__(self):
+        return self.wrapped_exc.__str__()
+
 
 # NOTE(salvatore-orlando): this class will go once the
 # extension API framework is updated
@@ -885,6 +1149,15 @@ class Controller(object):
 
     """
 
+    _view_builder_class = None
+
+    def __init__(self):
+        """Initialize controller with a view builder instance."""
+        if self._view_builder_class:
+            self._view_builder = self._view_builder_class()
+        else:
+            self._view_builder = None
+
     @webob.dec.wsgify(RequestClass=Request)
     def __call__(self, req):
         """Call the method specified in req.environ by RoutesMiddleware."""
